package com.swagger.rest.controllers

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import com.swagger.rest.models.*
import com.swagger.rest.models.Enum
import com.swagger.rest.repositories.*
import com.swagger.rest.services.*
import jakarta.servlet.http.HttpServletResponse
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.web.bind.annotation.*
import org.springframework.web.multipart.MultipartFile
import org.supercsv.io.CsvBeanWriter
import org.supercsv.io.ICsvBeanWriter
import org.supercsv.prefs.CsvPreference
import java.io.IOException
import java.text.SimpleDateFormat
import java.util.*
import kotlin.Any
import kotlin.Boolean
import kotlin.RuntimeException
import kotlin.String
import kotlin.Throws
import kotlin.Unit
import kotlin.arrayOf
import kotlin.toString


@RestController
@RequestMapping("/api/v1")
class MaintenanceController(
    private val projectRepository: ProjectRepository,
    private val userRepository: UserRepository,
    private val platformRepository: PlatformRepository,
    private val memberRepository: MemberRepository,
    private val bugRepository: BugRepository,
    private val commentRepository: CommentRepository,
    private val userService: UserService,
    private val projectService: ProjectService,
    private val platformService: PlatformService,
    private val memberService: MemberService,
    private val commentService: CommentService,
    private val bugService: BugService
) {

    @GetMapping("/maintenance")
    @Throws(IOException::class)
    fun exportToCSV(response: HttpServletResponse, @RequestParam tableName: String) {
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        val valid = memberRepository.validAdmin(userId.id.toString())
        val tableList = listOf("project", "platform", "user", "project_member", "bug", "comment")
//        if (valid == 0) {
//            ResponseEntity<Any>(HttpStatus.FORBIDDEN)
//        } else {
        if (!tableList.contains(tableName) || tableName.isEmpty()) {
            ResponseEntity<Any>(HttpStatus.BAD_REQUEST)
        } else {

            val list = when (tableName) {
                "user" -> userRepository.findAll()
                "project" -> projectRepository.findAll().map {
                    ProjectOutput(
                        id = it.id, name = it.name, description = it.description, owner = it.owner!!.id.toString()
                    )
                }

                "platform" -> platformRepository.findAll().map {
                    PlatformOutput(
                        id = it.id, name = it.name, project = it.project!!.id.toString()
                    )
                }

                "project_member" -> memberRepository.findAll().map {
                    MemberOutput(
                        id = it.id,
                        project = it.project!!.id.toString(),
                        user = it.user!!.id.toString(),
                        role = Enum.Member.values()[it.role].ordinal.toString()
                    )
                }

                "bug" -> bugRepository.findAll().map {
                    BugOutput(
                        id = it.id,
                        created = SimpleDateFormat("dd-MMM-yy HH:mm:ss").format(it.created),
                        description = it.description,
                        qc = it.qc!!.id.toString(),
                        dev = it.dev!!.id.toString(),
                        platform = it.platform!!.id.toString(),
                        goodday_url = it.goodday_url,
                        image_url = it.image_url,
                        level = Enum.Level.values()[it.level].ordinal.toString(),
                        status = Enum.Status.values()[it.status].ordinal.toString(),
                        dev_status = Enum.Dev_Status.values()[it.dev_status].ordinal.toString()
                    )
                }

                "comment" -> commentRepository.findAll().map {
                    CommentOutput(
                        id = it.id,
                        bug = it.bug!!.id.toString(),
                        creator = it.creator!!.id.toString(),
                        created = SimpleDateFormat("dd-MMM-yy HH:mm:ss").format(it.created),
                        content = it.content
                    )
                }

                else -> listOf()
            }
            val csvHeader = when (tableName) {
                "user" -> arrayOf("User ID", "Username", "Password", "Name")
                "project" -> arrayOf("Project ID", "Description", "Name", "Owner")
                "platform" -> arrayOf("Platform ID", "Name", "Project")
                "project_member" -> arrayOf("Member ID", "Role", "Project", "User")
                "comment" -> arrayOf("Bug ID", "Content", "Created", "Bug", "Creator")
                "bug" -> arrayOf(
                    "Bug ID",
                    "created",
                    "Description",
                    "Dev Status",
                    "Goodday Url",
                    "Image Url",
                    "Level",
                    "Status",
                    "Dev",
                    "Platform",
                    "Qc"
                )

                else -> arrayOf()
            }
            val nameMapping = when (tableName) {
                "user" -> arrayOf("id", "username", "password", "name")
                "project" -> arrayOf("id", "description", "name", "owner")
                "platform" -> arrayOf("id", "name", "project")
                "project_member" -> arrayOf("id", "role", "project", "user")
                "comment" -> arrayOf("id", "content", "created", "bug", "creator")
                "bug" -> arrayOf(
                    "id",
                    "created",
                    "description",
                    "dev_status",
                    "goodday_url",
                    "image_url",
                    "level",
                    "status",
                    "dev",
                    "platform",
                    "qc"
                )

                else -> arrayOf()
            }
            val csvWriter: ICsvBeanWriter = CsvBeanWriter(response.writer, CsvPreference.STANDARD_PREFERENCE)
            csvWriter.writeHeader(*csvHeader)
            for (table in list) {
                csvWriter.write(table, *nameMapping)
            }
            csvWriter.close()
//            }
        }
        //todo multi / zip download?
    }

    @ResponseStatus(HttpStatus.BAD_REQUEST)
    internal class Error : RuntimeException("oxes")

    @PostMapping("/maintenance")
    fun import(
        @RequestParam file: MultipartFile, @RequestParam tableName: String, @RequestParam overwrite: Boolean
    ): ResponseEntity<Any> {
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        val valid = memberRepository.validAdmin(userId.id.toString())
        val tableList = listOf("project", "platform", "user", "project_member", "bug", "comment")
//        if (valid == 0) {
//            ResponseEntity<Any>(HttpStatus.FORBIDDEN)
//        } else {
        var found = 0
        val mapData: MutableList<MutableMap<String, Any>> = mutableListOf()
        if (!tableList.contains(tableName) || tableName.isEmpty()) {
            ResponseEntity<Any>(HttpStatus.BAD_REQUEST)
        } else {
            val nameMapping = when (tableName) {
                "user" -> arrayOf("id", "username", "password", "name")
                "project" -> arrayOf("id", "description", "name", "owner")
                "platform" -> arrayOf("id", "name", "project")
                "project_member" -> arrayOf("id", "role", "project", "user")
                "comment" -> arrayOf("id", "content", "created", "bug", "creator")
                "bug" -> arrayOf(
                    "id",
                    "created",
                    "description",
                    "dev_status",
                    "goodday_url",
                    "image_url",
                    "level",
                    "status",
                    "dev",
                    "platform",
                    "qc"
                )

                else -> arrayOf()
            }
            val content = String(file.bytes)
            val lines = content.lines()
            lines.forEachIndexed { ind, t ->
                if (ind == 0 && arrayOf(t) !== nameMapping) {
                    ResponseEntity<Any>(HttpStatus.BAD_REQUEST)
                } else if (t !== "" && ind > 0) {
                    val commas = t.split(",(?=(?:[^\"]*\"[^\"]*\")*[^\"]*$)".toRegex()).toTypedArray()
                    val map: MutableMap<String, Any> = mutableMapOf()
                    nameMapping.forEachIndexed { index, s ->
                        map[s] = commas[index]
                    }
                    mapData.add(map)
                }
            }
            val existing = when (tableName) {
                "project" -> projectRepository.getId()
                "platform" -> platformRepository.getId()
                "user" -> userRepository.getId()
                "project_member" -> memberRepository.getId()
                "bug" -> bugRepository.getId()
                "comment" -> commentRepository.getId()
                else -> listOf()
            }
            mapData.removeIf { p -> existing.contains(p["id"].toString().toLong()) && !overwrite }

            val check = when (tableName) {
                "project" -> mapData.forEach {
                    val projectIn = ProjectInput()
                    projectIn.name = it["name"].toString()
                    projectIn.owner = it["owner"].toString().toLong()
                    projectIn.ownerUsername =
                        userRepository.findById(projectIn.owner!!.toLong()).orElseThrow { Error() }.username
                    projectIn.id = it["id"].toString().toLong()
                    projectIn.description = it["description"].toString()
                    val validate = projectService.validate(projectIn)
                    found += if (validate !== HttpStatus.OK && validate !== HttpStatus.CONFLICT) 1 else 0
                }

                "platform" -> mapData.forEach {
                    val platformInput = PlatformInput()
                    platformInput.id = it["id"].toString().toLong()
                    platformInput.name = it["name"].toString()
                    platformInput.project_id = it["project"].toString().toLong()
                    platformInput.project_name =
                        projectRepository.findById(platformInput.project_id!!.toLong()).orElseThrow { Error() }.name
                    val validate = platformService.validate(platformInput)
                    found += if (validate !== HttpStatus.OK && validate !== HttpStatus.CONFLICT && validate !== HttpStatus.FORBIDDEN) 1 else 0
                }

                "user" -> mapData.forEach {
                    val validate = userService.validate(
                        jacksonObjectMapper().readValue(
                            jacksonObjectMapper().writeValueAsString(it), UserInput::class.java
                        )
                    )
                    found += if (validate !== HttpStatus.OK && validate !== HttpStatus.CONFLICT) 1 else 0
                }

                "project_member" -> mapData.forEachIndexed { _, m ->
                    val memberInput = MemberInput()
                    memberInput.id = m["id"].toString().toLong()
                    memberInput.roleOrdinal = Integer.parseInt(m["role"].toString())
                    memberInput.role = if (Enum.Member.values()
                            .any { it.ordinal == memberInput.roleOrdinal }
                    ) Enum.Member.values()[memberInput.roleOrdinal!!].name else throw Error()
                    memberInput.user_id = m["user"].toString().toLong()
                    val validate = memberService.validate(null, memberInput)
                    found += if (validate !== HttpStatus.OK && validate !== HttpStatus.CONFLICT && validate !== HttpStatus.FORBIDDEN) 1 else 0
                }

                "comment" -> mapData.forEach {
                    val commentInput = CommentInput()
                    commentInput.id = it["id"].toString().toLong()
                    commentInput.bug_id = it["bug"].toString().toLong()
                    commentInput.bug = bugRepository.findById(commentInput.bug_id!!).orElseThrow { Error() }.description
                    commentInput.content = it["content"].toString()
                    commentInput.creator = it["creator"].toString()
                    commentInput.created = it["created"].toString()
                    val validate = commentService.validate(commentInput.bug_id, commentInput)
                    found += if (validate !== HttpStatus.OK) 1 else 0
                }

                "bug" -> mapData.forEachIndexed { _, m ->
                    val bugInput = BugInput()
                    bugInput.id = m["id"].toString().toLong()
                    bugInput.created = m["created"].toString()
                    bugInput.description = m["description"].toString()
                    bugInput.devStatOrdinal = Integer.parseInt(m["dev_status"].toString())
                    bugInput.dev_status = if (Enum.Dev_Status.values()
                            .any { it.ordinal == bugInput.devStatOrdinal }
                    ) Enum.Dev_Status.values()[bugInput.devStatOrdinal!!].name else throw Error()
                    bugInput.levelOrdinal = Integer.parseInt(m["level"].toString())
                    bugInput.level = if (Enum.Level.values()
                            .any { it.ordinal == bugInput.levelOrdinal }
                    ) Enum.Level.values()[bugInput.levelOrdinal!!].name else throw Error()
                    bugInput.statusOrdinal = Integer.parseInt(m["status"].toString())
                    bugInput.status = if (Enum.Status.values()
                            .any { it.ordinal == bugInput.statusOrdinal }
                    ) Enum.Status.values()[bugInput.statusOrdinal!!].name else throw Error()
                    bugInput.goodday_url = m["goodday_url"].toString()
                    bugInput.image_url = m["image_url"].toString()
                    bugInput.platformId = m["platform"].toString().toLong()
                    bugInput.platform = platformRepository.findById(bugInput.platformId).orElseThrow { Error() }.name
                    bugInput.dev = m["dev"].toString().toLong()
                    bugInput.qc = m["qc"].toString().toLong()
                    val validate = bugService.validate(bugInput)
                    found += if (validate !== HttpStatus.OK && validate !== HttpStatus.CONFLICT && validate !== HttpStatus.FORBIDDEN) 1 else 0
                }

                else -> Unit
            }
        }
        val project = Project()
        val platform = Platform()
        var user = User()
        val member = ProjectMember()
        val comment = Comment()
        val bug = Bug()
        if (found == 0) {
            val unit = when (tableName) {
                "project" -> mapData.forEach {
                    project.id = it["id"].toString().toLong()
                    project.name = it["name"].toString()
                    project.description = it["description"].toString()
                    project.owner = userRepository.findById(it["owner"].toString().toLong()).get()
                }

                "user" -> mapData.forEach {
                    user = jacksonObjectMapper().readValue(
                        jacksonObjectMapper().writeValueAsString(it), User::class.java
                    )
                }

                "platform" -> mapData.forEach {
                    platform.id = it["id"].toString().toLong()
                    platform.name = it["name"].toString()
                    platform.project = projectRepository.findById(it["project"].toString().toLong()).get()
                }

                "project_member" -> mapData.forEach {
                    member.id = it["id"].toString().toLong()
                    member.role = Enum.Member.values()[Integer.parseInt(it["role"].toString())].ordinal
                    member.project = projectRepository.findById(it["project"].toString().toLong()).get()
                    member.user = userRepository.findById(it["user"].toString().toLong()).get()
                }

                "comment" -> mapData.forEach {
                    comment.id = it["id"].toString().toLong()
                    comment.content = it["content"].toString()
                    comment.created =
                        SimpleDateFormat("dd-MMM-yy HH:mm:ss", Locale.ENGLISH).parse(it["created"].toString())
                    comment.bug = bugRepository.findById(it["bug"].toString().toLong()).get()
                    comment.creator = userRepository.findById(it["creator"].toString().toLong()).get()
                }

                "bug" -> mapData.forEach {
                    bug.id = it["id"].toString().toLong()
                    bug.created = SimpleDateFormat("dd-MMM-yy HH:mm:ss", Locale.ENGLISH).parse(it["created"].toString())
                    bug.description = it["description"].toString()
                    bug.dev_status = Enum.Dev_Status.values()[Integer.parseInt(it["dev_status"].toString())].ordinal
                    bug.goodday_url = it["goodday_url"].toString()
                    bug.image_url = it["image_url"].toString()
                    bug.level = Enum.Level.values()[Integer.parseInt(it["level"].toString())].ordinal
                    bug.status = Enum.Status.values()[Integer.parseInt(it["status"].toString())].ordinal
                    bug.dev = userRepository.findById(it["dev"].toString().toLong()).get()
                    bug.platform = platformRepository.findById(it["platform"].toString().toLong()).get()
                    bug.qc = userRepository.findById(it["qc"].toString().toLong()).get()
                }

                else -> Unit
            }
        }
        if (found == 0) {
            when (tableName) {
                "project" -> projectRepository.save(project)
                "platform" -> platformRepository.save(platform)
                "project_member" -> memberRepository.save(member)
                "user" -> userRepository.save(user)
                "comment" -> commentRepository.save(comment)
                "bug" -> bugRepository.save(bug)
                else -> Unit
            }
        }
        return if (found == 0) ResponseEntity<Any>(HttpStatus.OK) else ResponseEntity<Any>(HttpStatus.BAD_REQUEST)
    }
}


