package com.swagger.rest.controllers

import com.swagger.rest.models.Project
import com.swagger.rest.models.ProjectInput
import com.swagger.rest.models.ProjectMember
import com.swagger.rest.repositories.MemberRepository
import com.swagger.rest.repositories.PlatformRepository
import com.swagger.rest.repositories.ProjectRepository
import com.swagger.rest.repositories.UserRepository
import com.swagger.rest.services.ProjectService
import jakarta.persistence.criteria.Predicate
import org.springframework.data.domain.PageRequest
import org.springframework.data.domain.Pageable
import org.springframework.data.domain.Sort
import org.springframework.data.jpa.domain.Specification
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.web.bind.annotation.*
import java.util.*

@RestController
@RequestMapping("/api/v1")
class ProjectController(
    private val projectRepository: ProjectRepository,
    private val platformRepository: PlatformRepository,
    private val userRepository: UserRepository,
    private val memberRepository: MemberRepository,
    private val projectService: ProjectService
) {
    private fun getSortDirection(direction: String): Sort.Direction {
        if (direction == "asc") {
            return Sort.Direction.ASC
        } else if (direction == "desc") {
            return Sort.Direction.DESC
        }
        return Sort.Direction.ASC
    }

    @GetMapping("/projects")
    fun getProject(
        @RequestParam(defaultValue = 0.toString()) page: Int,
        @RequestParam(defaultValue = 3.toString()) limit: Int,
        @RequestParam(defaultValue = "project_id, desc") sort: Array<String>
    ): Any {
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        return try {
            val orders: MutableList<Sort.Order> = ArrayList()
            val column = listOf("project_id", "name", "description", "owner")
            val sort2 = if (!sort.contains(",")) {
                sort + ",desc"
            } else {
                sort
            }
            if (!column.contains(sort2[0])) {
                ResponseEntity<Project>(HttpStatus.BAD_REQUEST)
            } else {
                if (sort2[0].contains(",")) {
                    // will sort more than 2 fields
                    // sortOrder="field, direction"
                    for (sortOrder in sort2) {
                        val _sort = sortOrder.split(",".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
                        orders.add(Sort.Order(getSortDirection(_sort[1]), _sort[0]))
                    }
                } else {
                    // sort=[field, direction]
                    orders.add(Sort.Order(getSortDirection(sort2[1]), sort2[0]))
                }
                val pagingSort: Pageable = PageRequest.of(page, limit, Sort.by(orders))
                val projects = projectRepository.findOwnerOrMember(userId.id.toString(), pagingSort)
                val ret = projects.content
                val response: MutableMap<String, Any> = HashMap()
                response["currentPage"] = projects.number
                response["totalRecord"] = projects.totalElements
                response["totalPage"] = projects.totalPages
                response["results"] = ret
                if (ret.isEmpty()) {
                    arrayOf<String>()
                } else {
                    ResponseEntity(response, HttpStatus.OK)
                }
            }
        } catch (e: Exception) {
            e.printStackTrace()
            ResponseEntity(e, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @GetMapping("/projects/{id}")
    fun getProjectById(@PathVariable("id") id: Long): ResponseEntity<Project> {
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        val validOwner = projectRepository.validOwner(id.toString(), userId.id.toString())
        val validMember = memberRepository.validMember(id.toString(), userId.id.toString())
        val projectData: Optional<Project?> = projectRepository.findById(id)
        return if (projectData.isEmpty) {
            ResponseEntity<Project>(HttpStatus.NOT_FOUND)
        } else {
            if (validOwner > 0 || validMember > 0) {
                ResponseEntity<Project>(projectData.get(), HttpStatus.OK)
            } else {
                ResponseEntity<Project>(HttpStatus.FORBIDDEN)
            }
        }
    }

    @PostMapping("/projects")
    fun addProject(@RequestBody project: ProjectInput): ResponseEntity<Project> {
        val valid = projectService.validate(project)
        return if (valid !== HttpStatus.OK) {
            ResponseEntity<Project>(valid!!)
        } else {
            val saveProject = projectService.saveProject(project)
            ResponseEntity<Project>(saveProject, HttpStatus.CREATED)
        }
    }

    @PutMapping("/projects/{id}")
    fun updateProjectById(@PathVariable("id") id: Long, @RequestBody project: ProjectInput): ResponseEntity<out Any?> {
        val projectData = projectRepository.findById(id)
        val found = projectRepository.countByName(project.name)
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        val validOwner = projectRepository.validOwner(id.toString(), userId.id.toString())
        val validAdmin = memberRepository.validRole(id.toString(), userId.id.toString(), "2")
        return if (projectData.isPresent) {
            if (validOwner > 0 || validAdmin > 0) {
                if (project.name.isNotBlank()) {
                    if (project.name.length > 100 || project.description!!.length > 255) {//too long
                        ResponseEntity<Project>(HttpStatus.PAYLOAD_TOO_LARGE)
                    } else {
                        if ((project.name == projectData.get().name && found > 0) || (project.name !== projectData.get().name && found == 0)) {
                            val saveProject = projectService.updateProject(id, project)
                            ResponseEntity<Any>(saveProject, HttpStatus.OK)
                        } else {//duplicate
                            ResponseEntity<Project>(HttpStatus.CONFLICT)
                        }
                    }
                } else {//name invalid
                    ResponseEntity<Project?>(HttpStatus.BAD_REQUEST)
                }
            } else {
                ResponseEntity<Project>(HttpStatus.FORBIDDEN)
            }
        } else {
            ResponseEntity<Project?>(HttpStatus.NOT_FOUND)
        }
    }

    @DeleteMapping("/projects/{id}")
    fun deleteProject(@PathVariable("id") id: Long): ResponseEntity<HttpStatus> {
        val find = projectRepository.findById(id)
        val used = platformRepository.findByProjectId(id.toString()).size
        val spec = Specification<ProjectMember> { root, _, builder ->
            val list: MutableList<Predicate> = mutableListOf()
            list.add(builder.equal(root.get<Project>("project").get<Long>("id"), id))
            builder.and(*list.toTypedArray())
        }
        val usedMember = memberRepository.count(spec)
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        val validOwner = projectRepository.validOwner(id.toString(), userId.id.toString())
        return try {
            if (validOwner > 0) {
                if (used > 0 || usedMember > 0) {//child used in transaction
                    ResponseEntity(HttpStatus.RESET_CONTENT)
                } else if (find.isPresent) {
                    projectRepository.deleteById(id)
                    ResponseEntity(HttpStatus.OK)
                } else {//project not found
                    ResponseEntity(HttpStatus.NOT_FOUND)
                }
            } else {
                ResponseEntity(HttpStatus.FORBIDDEN)
            }
        } catch (e: Exception) {
            ResponseEntity(HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

}